/*
 * Copyright 2020 Tadashi G. Takaoka
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __LIBASM_DIS_MOS6502_H__
#define __LIBASM_DIS_MOS6502_H__

#include "config_mos6502.h"
#include "dis_base.h"
#include "insn_mos6502.h"

namespace libasm {
namespace mos6502 {

struct DisMos6502 final : Disassembler, Config {
    DisMos6502(const ValueFormatter::Plugins &plugins = defaultPlugins());

    void reset() override;

    Error setLongAccumulator(bool enable);
    Error setLongIndex(bool enable);
    Error setUseIndirectLong(bool enable);

private:
    const BoolOption<DisMos6502> _opt_longa;
    const BoolOption<DisMos6502> _opt_longi;
    const BoolOption<DisMos6502> _opt_indirectLong;

    bool _longAccumulator;
    bool _longIndex;
    bool _useIndirectLong;
    bool longImmediate(AddrMode mode) const;

    void decodeImmediate(DisInsn &insn, StrBuffer &out, AddrMode mode) const;
    void decodeAbsoluteLong(DisInsn &insn, StrBuffer &out) const;
    void decodeAbsolute(DisInsn &insn, StrBuffer &out) const;
    void decodeDirectPage(DisInsn &insn, StrBuffer &out) const;
    void decodeRelative(DisInsn &insn, StrBuffer &out, AddrMode mode) const;
    void decodeBlockMove(DisInsn &insn, StrBuffer &out) const;
    void decodeOperand(DisInsn &insn, StrBuffer &out, AddrMode mode) const;

    Error decodeImpl(DisMemory &memory, Insn &insn, StrBuffer &out) const override;
    const ConfigBase &config() const override { return *this; }
    ConfigSetter &configSetter() override { return *this; }
    static const ValueFormatter::Plugins &defaultPlugins();
};

}  // namespace mos6502
}  // namespace libasm

#endif  // __LIBASM_DIS_MOS6502_H__

// Local Variables:
// mode: c++
// c-basic-offset: 4
// tab-width: 4
// End:
// vim: set ft=cpp et ts=4 sw=4:
